Search.setIndex({"alltitles": {"Indices and tables": [[0, "indices-and-tables"]], "Welcome to ChaProEV\u2019s documentation!": [[0, "welcome-to-chaproev-s-documentation"]]}, "docnames": ["index"], "envversion": {"sphinx": 61, "sphinx.domains.c": 3, "sphinx.domains.changeset": 1, "sphinx.domains.citation": 1, "sphinx.domains.cpp": 9, "sphinx.domains.index": 1, "sphinx.domains.javascript": 3, "sphinx.domains.math": 2, "sphinx.domains.python": 4, "sphinx.domains.rst": 2, "sphinx.domains.std": 2}, "filenames": ["index.rst"], "indexentries": {}, "objects": {}, "objnames": {}, "objtypes": {}, "terms": {"index": [], "modul": [], "page": 0, "search": 0}, "titles": ["Welcome to ChaProEV\u2019s documentation!"], "titleterms": {"": 0, "chaproev": 0, "document": 0, "indic": 0, "tabl": 0, "welcom": 0}})